package com.sw.mes.store.api.controller;

import com.alibaba.fastjson.JSON;
import com.baomidou.mybatisplus.core.toolkit.ObjectUtils;
import com.sw.mes.core.dto.store.WarehouseInputBillBo;
import com.sw.mes.core.dto.store.WarehouseMaterialBo;
import com.sw.mes.store.dto.*;
import com.sw.mes.core.dto.store.WarehouseOutputBillBo;
import com.sw.mes.core.enums.MessageEnum;
import com.sw.mes.core.response.Result;
import com.sw.mes.core.response.ResultGenerator;
import com.sw.mes.store.api.interfaces.supplier.WarehouseSupplierService;
import com.sw.mes.store.entity.WarehouseInventory;
import com.sw.mes.store.entity.WarehouseModel;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import io.swagger.annotations.ApiParam;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.web.bind.annotation.*;

import javax.validation.Valid;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * @Description 提供给外部系统调用的接口
 * @Author wupeng
 * @Date 2023/12/28 17:25
 **/
@RestController
@RequestMapping("/warehouse/supplier")
@Slf4j
@Api(tags = "线边仓提供给外部系统调用的接口")
public class WarehouseSupplierController {

    @Autowired
    private WarehouseSupplierService warehouseSupplierService;

    /**
     * 入库申请接口
     **/
    @ApiOperation(value = "入库申请")
    @PostMapping("/inputApply")
    public Result<String> inputApply(@RequestBody @Valid @ApiParam(value = "入库单对象") WarehouseInputBillBo inputBillBo) {
        try {
            String billNo = warehouseSupplierService.input(inputBillBo);
            return ResultGenerator.success(billNo);
        } catch (Exception e) {
            return ResultGenerator.fail(e.getMessage());
        }
    }

    /**
     * 入库申请接口
     **/
    @ApiOperation(value = "入库申请")
    @PostMapping("/inputApplyData")
    public Result<WarehouseInputReturnDto> inputApplyData(@RequestBody @Valid @ApiParam(value = "入库单对象") WarehouseInputBillBo inputBillBo) {
        try {
            WarehouseInputReturnDto warehouseInputReturnDto = warehouseSupplierService.inputData(inputBillBo);
            return ResultGenerator.success(warehouseInputReturnDto);
        } catch (Exception e) {
            return ResultGenerator.fail(e.getMessage());
        }
    }

    /**
     * 取消入库申请接口
     **/
    @ApiOperation(value = "取消入库申请")
    @PostMapping("/cancelInputApply")
    public Result cancelInputApply(@RequestParam("billNo") @ApiParam(value = "入库单号") String billNo) {
        try {
            warehouseSupplierService.cancelInput(billNo);
        } catch (Exception e) {
            return ResultGenerator.fail(e.getMessage());
        }
        return ResultGenerator.success();
    }

    /**
     * 出库申请接口
     **/
    @ApiOperation(value = "出库申请")
    @PostMapping("/outputApply")
    public Result<String> outputApply(@RequestBody @Valid @ApiParam(value = "出库单对象") WarehouseOutputBillBo outputBillBo) {
        try {
            String billNo = warehouseSupplierService.output(outputBillBo);
            return ResultGenerator.success(billNo);
        } catch (Exception e) {
            return ResultGenerator.fail(e.getMessage());
        }
    }
    @ApiOperation(value = "出库申请返回对象")
    @PostMapping("/outputApplyData")
    public Result<WarehouseInputReturnDto> outputApplyData(@RequestBody @Valid @ApiParam(value = "出库单对象") WarehouseOutputBillBo outputBillBo) {
        try {
            WarehouseInputReturnDto warehouseInputReturnDto = warehouseSupplierService.outputData(outputBillBo);
            return ResultGenerator.success(warehouseInputReturnDto);
        } catch (Exception e) {
            return ResultGenerator.fail(e.getMessage());
        }
    }
    /**
     * 取消入库申请接口
     **/
    @ApiOperation(value = "取消出库申请")
    @PostMapping("/cancelOutputApply")
    public Result cancelOutputApply(@RequestParam("billNo") @ApiParam(value = "出库单号") String billNo) {
        try {
            warehouseSupplierService.cancelOutput(billNo);
        } catch (Exception e) {
            return ResultGenerator.fail(e.getMessage());
        }
        return ResultGenerator.success();
    }
    @ApiOperation(value = "创建入库单AND调拨单AND出库单接口")
    @PostMapping("/createInputAndAllotAndOut")
    Result createInputAndAllotAndOut(@RequestBody @Valid StoreDto storeDto){
        warehouseSupplierService.createInputAndAllotAndOut(storeDto);
        return ResultGenerator.success();
    }
    /**
     * 通过批次号获取库存信息
     */
    @ApiOperation(value = "通过批次号获取库存信息")
    @GetMapping("/findByBatchNo")
    Result<WarehouseInventory> findByBatchNo(@RequestParam("batchNo") @ApiParam(value = "批次号") String batchNo) {
        try {
            List<WarehouseInventory> inventories = warehouseSupplierService.findByBatchNo(batchNo);
            if(ObjectUtils.isEmpty(inventories)){
                return  ResultGenerator.fail("未查询到数据!");
            }
            //所有批次号对应的主数据料号都应该是一样的话返回一条数据即可，否则返回错误信息
            Map<String, List<WarehouseInventory>> inventoriesGroups = inventories.stream()
                    .collect(Collectors.groupingBy(WarehouseInventory::getMaterialModelCode));
            if(inventoriesGroups.size()==1){
                return ResultGenerator.success(MessageEnum.SUCCESS.getMessage(),inventories.get(0));
            }else {
                return ResultGenerator.fail("查询数据存在错误数据!");
            }
        } catch (Exception e) {
            return ResultGenerator.fail(e.getMessage());
        }
    }

    @ApiOperation(value = "对外创建调拨单接口")
    @PostMapping("/createAllotBill")
    public Result createAllotBill(@RequestBody @Valid @ApiParam(value = "调拨单对象") WarehouseAllotBillDto warehouseAllotBillDto) {
        Boolean allotBill = warehouseSupplierService.createAllotBill(warehouseAllotBillDto);
        if(allotBill){
            return ResultGenerator.success("调拨单创建成功!",null);
        }else{
            return ResultGenerator.fail("调拨单创建失败!");
        }
    }
    /**
     * 获取当前节点的所有子节点id
     * @param id
     * @return
     */
    @GetMapping("/getChildrenId")
    @ApiOperation("获取当前节点的所有子节点id")
    public Result<List<Long>> getByWarehouseModelId(@RequestParam("id")Long id) {
        List<WarehouseModelDto> storeLocations = warehouseSupplierService.getChildLocationIds(id);
        List<Long> result = storeLocations.stream().map(storeLocationId -> {
            return storeLocationId.getId();
        }).collect(Collectors.toList());
        return ResultGenerator.success(result);
    }

    @ApiOperation(value = "对外申请入库且是否自动出库接口")
    @PostMapping("/automaticInputOut")
    public Result AutomaticInputOut(@RequestBody @Valid @ApiParam(value = "申请入库单对象") WarehouseInputBillBo inputBillBo) {
        Result result = warehouseSupplierService.automaticInputOut(inputBillBo);
        if(result.getCode().equals(200)){
            return ResultGenerator.success("生成入库单成功!",result.getData());
        }else{
            return ResultGenerator.fail("生成入库单成功!");
        }
    }

    @ApiOperation(value = "根据批次号and库位获取库存信息")
    @GetMapping("/getByBatchWarehouseId")
    Result getByBatchWarehouseId(@RequestParam String batchNo, @RequestParam Long warehouseId){
        WarehouseInventoryDto byBatchWarehouseId = warehouseSupplierService.getByBatchWarehouseId(batchNo, warehouseId);
        return ResultGenerator.success(byBatchWarehouseId);
    }

    @ApiOperation(value = "线边仓模型详细信息")
    @GetMapping("/getById")
    public Result<WarehouseModelDto> getById(@RequestParam @ApiParam(value = "线边仓模型id") Long id) {
        WarehouseModel warehouseModel = warehouseSupplierService.getById(id);
        if (null == warehouseModel) {
            return ResultGenerator.fail("线边仓模型不存在");
        }
        WarehouseModelDto warehouseModelDto = new WarehouseModelDto();
        BeanUtils.copyProperties(warehouseModel,warehouseModelDto);
        return ResultGenerator.success(warehouseModelDto);
    }

    @ApiOperation(value = "同步物料主数据")
    @PostMapping("/syncMaterial")
    public Result syncMaterial(@RequestBody List<WarehouseMaterialBo> warehouseMaterialBos) {
        Result result = warehouseSupplierService.syncMaterial(warehouseMaterialBos);
        if(result.getCode().equals(200)){
            return ResultGenerator.success("同步物料主数据成功!",null);
        }else{
            return ResultGenerator.fail("同步物料主数据失败!");
        }
    }
}
