package com.sw.mes.store.controller;

import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.core.metadata.OrderItem;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.sw.mes.core.response.ResponsePage;
import com.sw.mes.core.response.Result;
import com.sw.mes.core.response.ResultGenerator;
import com.sw.mes.core.utils.DynmaicQueryWrapper;
import com.sw.mes.core.utils.RequestPage;
import com.sw.mes.store.dto.WarehouseOutputBillDto;
import com.sw.mes.store.dto.WmsOutputBillDto;
import com.sw.mes.store.entity.WarehouseOutputBill;
import com.sw.mes.store.entity.WarehouseOutputBillDetail;
import com.sw.mes.store.service.WarehouseOutputBillDetailService;
import com.sw.mes.store.service.WarehouseOutputBillService;
import io.swagger.annotations.*;
import lombok.extern.slf4j.Slf4j;
import org.modelmapper.ModelMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.validation.BindingResult;
import org.springframework.web.bind.annotation.*;
import springfox.documentation.annotations.ApiIgnore;

import javax.servlet.ServletRequest;
import javax.validation.Valid;
import java.util.List;
import java.util.stream.Collectors;

/**
 * @Author wupeng
 * @Description 出库单
 * @Date 18:44 2023/7/31
 * @Param
 * @return
 **/
@RestController
@RequestMapping("/warehouse/outputBill")
@Slf4j
@Api(tags = "线边仓-出库单")
public class WarehouseOutputBillController {
    @Autowired
    ModelMapper modelMapper;

    @Autowired
    private WarehouseOutputBillService warehouseOutputBillService;

    @Autowired
    private WarehouseOutputBillDetailService detailService;

    @ApiOperation(value = "分页查询出库单列表")
    @ApiImplicitParams({
            @ApiImplicitParam(name = "rows", value = "行数", dataType = "integer", paramType = "query"),
            @ApiImplicitParam(name = "page", value = "页码", dataType = "integer", paramType = "query"),
    })
    @GetMapping("/listByPage")
    public Result<ResponsePage<WarehouseOutputBillDto>> listByPage(ServletRequest request, @ApiIgnore RequestPage requestPage) {
        QueryWrapper<WarehouseOutputBill> queryWrapper = DynmaicQueryWrapper.bySearchFilter(request, WarehouseOutputBill.class);
        Page<WarehouseOutputBill> page = new Page<>(requestPage.getPage(), requestPage.getRows());
        String orderDirection = requestPage.getOrderDirection();
        String orderField = requestPage.getOrderField();
        page.addOrder(RequestPage.ORDER_DIRECTION_DESC.equals(orderDirection) ? OrderItem.desc(orderField) : OrderItem.asc(orderField));
        Page<WarehouseOutputBill> listPage = warehouseOutputBillService.page(page, queryWrapper);
        List<WarehouseOutputBillDto> dtos = listPage.getRecords().stream()
                .map(entity -> toLazyDto(entity))
                .collect(Collectors.toList());
        ResponsePage<WarehouseOutputBillDto> responsePage = new ResponsePage(dtos, listPage.getTotal());
        return ResultGenerator.success(responsePage);
    }

    @ApiOperation(value = "出库单详细信息")
    @GetMapping("/{id}")
    public Result<WarehouseOutputBillDto> get(@PathVariable @ApiParam(value = "出库单id") Long id) {
        WarehouseOutputBill warehouseOutputBill = warehouseOutputBillService.getById(id);
        if (null == warehouseOutputBill) {
            return ResultGenerator.fail("该出库单不存在");
        }
        List<WarehouseOutputBillDetail> outputBillDetails = detailService.getByBillId(id);
        warehouseOutputBill.setOutputBillDetails(outputBillDetails);
        return ResultGenerator.success(toEagerDto(warehouseOutputBill));
    }

    @ApiOperation(value = "人工创建出库单")
    @PostMapping("/create")
    public Result<String> create(@RequestBody @Valid @ApiParam(value = "出库单对象") WarehouseOutputBillDto warehouseOutputBillDto, BindingResult bindingResult) {
        if (bindingResult.hasErrors()) {
            return ResultGenerator.fail(bindingResult.getFieldError().getDefaultMessage());
        }
        try {
            WarehouseOutputBill warehouseOutputBill = fromDto(warehouseOutputBillDto);
            //保存入库单和入库明细
            warehouseOutputBillService.saveOutputBill(warehouseOutputBill);
            return ResultGenerator.success();
        } catch (Exception e) {
            return ResultGenerator.fail(e.getMessage());
        }
    }

    @ApiOperation(value = "确认出库")
    @PostMapping("/deduction")
    public Result<String> output(@RequestBody @Valid @ApiParam(value = "出库单对象") WarehouseOutputBillDto warehouseOutputBillDto, BindingResult bindingResult) {
        if (bindingResult.hasErrors()) {
            return ResultGenerator.fail(bindingResult.getFieldError().getDefaultMessage());
        }
        try {
            WarehouseOutputBill warehouseOutputBill = fromDto(warehouseOutputBillDto);
            //审核通过，扣减库存
            warehouseOutputBillService.auditDeductionAtomic(warehouseOutputBill);
            return ResultGenerator.success();
        } catch (Exception e) {
            return ResultGenerator.fail(e.getMessage());
        }
    }

    @ApiOperation(value = "取消出库单")
    @PostMapping("/cancel")
    public Result<String> cancel(@RequestBody @Valid @ApiParam(value = "出库单对象") WarehouseOutputBillDto warehouseOutputBillDto, BindingResult bindingResult) {
        if (bindingResult.hasErrors()) {
            return ResultGenerator.fail(bindingResult.getFieldError().getDefaultMessage());
        }
        try {
            WarehouseOutputBill warehouseOutputBill = fromDto(warehouseOutputBillDto);
            //取消出库单
            warehouseOutputBillService.cancelApply(warehouseOutputBill.getBillNo());
            return ResultGenerator.success();
        } catch (Exception e) {
            return ResultGenerator.fail(e.getMessage());
        }
    }
    @ApiOperation(value = "退料出库接收WMS状态反馈信息 ")
    @PostMapping("/statusReception")
    public Result statusReception(@RequestBody @Valid @ApiParam(value = "接收对象") WmsOutputBillDto dto) {
        try {
            warehouseOutputBillService.statusReception(dto);
            return ResultGenerator.success();
        } catch (Exception e) {
            return ResultGenerator.fail(e.getMessage());
        }
    }
    @ApiOperation(value = "拒绝出库")
    @PostMapping("/rejection")
    public Result<String> rejectApply(@RequestBody @Valid @ApiParam(value = "出库单对象") WarehouseOutputBillDto warehouseOutputBillDto, BindingResult bindingResult) {
        if (bindingResult.hasErrors()) {
            return ResultGenerator.fail(bindingResult.getFieldError().getDefaultMessage());
        }
        try {
            WarehouseOutputBill warehouseOutputBill = fromDto(warehouseOutputBillDto);
            //拒绝出库
            warehouseOutputBillService.rejectApply(warehouseOutputBill);
            return ResultGenerator.success();
        } catch (Exception e) {
            return ResultGenerator.fail(e.getMessage());
        }
    }

    /**
     * 场景：直接出库，无需审核
     **/
    @ApiOperation(value = "直接出库")
    @PostMapping("/direct")
    public Result direct(@RequestBody @Valid @ApiParam(value = "出库单对象") WarehouseOutputBillDto warehouseOutputBillDto, BindingResult bindingResult) {
        if (bindingResult.hasErrors()) {
            return ResultGenerator.fail(bindingResult.getFieldError().getDefaultMessage());
        }
        try {
            WarehouseOutputBill warehouseOutputBill = fromDto(warehouseOutputBillDto);
            warehouseOutputBill.setAutoAudit(true);
            warehouseOutputBillService.saveOutputBill(warehouseOutputBill);
            return ResultGenerator.success();
        } catch (Exception e) {
            return ResultGenerator.fail(e.getMessage());
        }
    }

    public WarehouseOutputBillDto toLazyDto(WarehouseOutputBill entity) {
        return modelMapper.map(entity, WarehouseOutputBillDto.class);
    }

    public WarehouseOutputBillDto toEagerDto(WarehouseOutputBill entity) {
        WarehouseOutputBillDto dto = modelMapper.map(entity, WarehouseOutputBillDto.class);
        return dto;
    }

    public WarehouseOutputBill fromDto(WarehouseOutputBillDto dto) {
        return modelMapper.map(dto, WarehouseOutputBill.class);
    }


}