package com.sw.mes.store.service.impl;


import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.sw.mes.core.context.BaseContextHandler;
import com.sw.mes.core.exception.BusinessException;
import com.sw.mes.redis.server.RedissonUtils;
import com.sw.mes.store.api.interfaces.consumer.WarehouseConsumerService;
import com.sw.mes.store.constants.WarehouseConstant;
import com.sw.mes.store.dto.WmsOutputBillDto;
import com.sw.mes.store.entity.*;
import com.sw.mes.store.enums.InputBillStatusEnum;
import com.sw.mes.store.enums.OutputBillStatusEnum;
import com.sw.mes.store.enums.OutputBillTypeEnum;
import com.sw.mes.store.mapper.WarehouseOutputBillMapper;
import com.sw.mes.store.service.*;
import com.sw.mes.store.utils.CodeGenUtil;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.ObjectUtils;
import org.apache.commons.lang3.StringUtils;
import org.modelmapper.ModelMapper;
import org.redisson.api.RLock;
import org.redisson.api.RedissonClient;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.math.BigDecimal;
import java.util.*;
import java.util.stream.Collectors;


/**
 * @Author wupeng
 * @Description 出库单
 * @Date 20:35 2023/8/2
 * @Param
 * @return
 **/
@Service
public class WarehouseOutputBillServiceImpl extends ServiceImpl<WarehouseOutputBillMapper, WarehouseOutputBill> implements WarehouseOutputBillService {
    private static final Logger LOGGER = LoggerFactory.getLogger(WarehouseOutputBillServiceImpl.class);

    @Autowired
    private ModelMapper modelMapper;

    @Autowired
    private RedissonClient redissonClient;

    @Autowired
    private WarehouseOutputBillMapper outputBillMapper;

    @Autowired
    private WarehouseOutputBillService warehouseOutputBillService;

    @Autowired
    private WarehouseOutputBillDetailService detailService;

    @Autowired
    private WarehouseInventoryService inventoryService;

    @Autowired
    private WarehouseConsumerService warehouseConsumerService;

    @Autowired
    private WarehouseOutputRecordBillService outputRecordBillService;

    @Autowired
    private WarehouseOutputRecordBillDetailService outputRecordBillDetailService;

    /**
     * 保存出库单及其明细，待出库状态
     **/
    @Override
    @Transactional(rollbackFor = Exception.class)
    public void saveOutputBill(WarehouseOutputBill outputBill) {
        List<WarehouseOutputBillDetail> outputBillDetails = outputBill.getOutputBillDetails();
        if (CollectionUtils.isEmpty(outputBillDetails)) {
            throw new BusinessException("出库明细为空，请检查数据！");
        }
        //出库单号
        String billNo = outputBill.getBillNo();
        billNo = StringUtils.isNotBlank(billNo) ? billNo : CodeGenUtil.genOutputBillNo();
        outputBill.setBillNo(billNo);
        outputBill.setStatus(OutputBillStatusEnum.待出库.getValue());
        if (ObjectUtils.isEmpty(outputBill.getCreateTime())) {
            outputBill.setCreateTime(new Date());
        }
        if (ObjectUtils.isEmpty(outputBill.getCreateUserName())) {
            outputBill.setCreateUserName(BaseContextHandler.getUsername());
        }
        if (ObjectUtils.isEmpty(outputBill.getCreateUserId())) {
            outputBill.setCreateUserId(BaseContextHandler.getUserID());
        }
        warehouseOutputBillService.saveOrUpdate(outputBill);

        //当开启了唯一码，用于判断是否存在相同的唯一码，只要唯一码不同，其他字段可以相同
        Set<String> uniqueCodeSet = new HashSet<>();

        //保存明细
        for (WarehouseOutputBillDetail outputBillDetail : outputBillDetails) {
            String uniqueCode = outputBillDetail.getUniqueCode();
            if (StringUtils.isNotBlank(uniqueCode)) {
                //校验当前即将出库的明细中唯一码是否重复
                if (!uniqueCodeSet.add(uniqueCode)) {
                    throw new BusinessException("唯一码【" + uniqueCode + "】重复，请检查数据！");
                }
            }
            outputBillDetail.setOutputBillId(outputBill.getId());
            outputBillDetail.setOutputBillNo(outputBill.getBillNo());
            if (ObjectUtils.isEmpty(outputBillDetail.getCreateTime())) {
                outputBillDetail.setCreateTime(new Date());
            }
            if (ObjectUtils.isEmpty(outputBillDetail.getCreateUserName())) {
                outputBillDetail.setCreateUserName(BaseContextHandler.getUsername());
            }
            if (ObjectUtils.isEmpty(outputBillDetail.getCreateUserId())) {
                outputBillDetail.setCreateUserId(BaseContextHandler.getUserID());
            }
            outputBillDetail.setType(outputBill.getType());
            outputBillDetail.setDestination(outputBill.getDestination());
            outputBillDetail.setStatus(OutputBillStatusEnum.待出库.getValue());
        }
        detailService.saveOrUpdateBatch(outputBillDetails);
        //如果是自动出库，无需审核，直接扣减库存
        if (outputBill.getAutoAudit()) {
            warehouseOutputBillService.autoDeductionAtomic(outputBill);
        }
    }
    @Override
    @Transactional(rollbackFor = Exception.class)
    public String saveOutputBills(WarehouseOutputBill outputBill) {
        List<WarehouseOutputBillDetail> outputBillDetails = outputBill.getOutputBillDetails();
        if (CollectionUtils.isEmpty(outputBillDetails)) {
            throw new BusinessException("出库明细为空，请检查数据！");
        }
        //出库单号
        String billNo = outputBill.getBillNo();
        billNo = StringUtils.isNotBlank(billNo) ? billNo : CodeGenUtil.genOutputBillNo();
        outputBill.setBillNo(billNo);
        outputBill.setStatus(OutputBillStatusEnum.待出库.getValue());
        if (ObjectUtils.isEmpty(outputBill.getCreateTime())) {
            outputBill.setCreateTime(new Date());
        }
        if (ObjectUtils.isEmpty(outputBill.getCreateUserName())) {
            outputBill.setCreateUserName(BaseContextHandler.getUsername());
        }
        if (ObjectUtils.isEmpty(outputBill.getCreateUserId())) {
            outputBill.setCreateUserId(BaseContextHandler.getUserID());
        }
        warehouseOutputBillService.saveOrUpdate(outputBill);

        //当开启了唯一码，用于判断是否存在相同的唯一码，只要唯一码不同，其他字段可以相同
        Set<String> uniqueCodeSet = new HashSet<>();

        //保存明细
        for (WarehouseOutputBillDetail outputBillDetail : outputBillDetails) {
            String uniqueCode = outputBillDetail.getUniqueCode();
            if (StringUtils.isNotBlank(uniqueCode)) {
                //校验当前即将出库的明细中唯一码是否重复
                if (!uniqueCodeSet.add(uniqueCode)) {
                    throw new BusinessException("唯一码【" + uniqueCode + "】重复，请检查数据！");
                }
            }
            outputBillDetail.setOutputBillId(outputBill.getId());
            outputBillDetail.setOutputBillNo(outputBill.getBillNo());
            if (ObjectUtils.isEmpty(outputBillDetail.getCreateTime())) {
                outputBillDetail.setCreateTime(new Date());
            }
            if (ObjectUtils.isEmpty(outputBillDetail.getCreateUserName())) {
                outputBillDetail.setCreateUserName(BaseContextHandler.getUsername());
            }
            if (ObjectUtils.isEmpty(outputBillDetail.getCreateUserId())) {
                outputBillDetail.setCreateUserId(BaseContextHandler.getUserID());
            }
            outputBillDetail.setType(outputBill.getType());
            outputBillDetail.setDestination(outputBill.getDestination());
            outputBillDetail.setStatus(OutputBillStatusEnum.待出库.getValue());
        }
        detailService.saveOrUpdateBatch(outputBillDetails);
        //如果是自动出库，无需审核，直接扣减库存
        if (outputBill.getAutoAudit()) {
            warehouseOutputBillService.autoDeductionAtomic(outputBill);
        }
        return billNo;
    }
    /**
     * 出库审核通过，扣减库存
     **/
    @Override
    @Transactional(rollbackFor = Exception.class)
    public void auditDeduction(WarehouseOutputBill outputBill) {
        WarehouseOutputBill byBillNo = this.getByBillNo(outputBill.getBillNo());
        if (byBillNo == null) {
            throw new BusinessException("出库单号【" + outputBill.getBillNo() + "】信息为空，请检查数据！");
        }
        String status = byBillNo.getStatus();
        if (!OutputBillStatusEnum.待出库.getValue().equals(status)) {
            throw new BusinessException("出库单号【" + outputBill.getBillNo() + "】状态不是【待出库】状态，审核失败！");
        }
        if(!outputBill.getType().equals(OutputBillTypeEnum.WMS退料)){
            auditDeductionInventory(outputBill);
            //TODO wupeng 此处需要调用其他需要反馈的系统接口
            warehouseConsumerService.afterAuditOutput(outputBill);
        }else {
            //WMS退料需要等返回成功之后才会操作库存
            warehouseConsumerService.afterAuditOutput(outputBill);
        }

    }

    @Override
    public void statusReception(WmsOutputBillDto dto) {
        if(dto.getStatus().equals("接收")){
            WarehouseOutputBill outputBill = outputBillMapper.getByBillNo(dto.getBillNo());
            List<WarehouseOutputBillDetail> outputBillDetails = new ArrayList<>();
            dto.getDetailsDtos().stream().forEach(entity->{
                WarehouseOutputBillDetail detail = detailService.getByBillMaterialCode(dto.getBillNo(),entity.getSku());
                detail.setOutputCount(entity.getNumber());
                outputBillDetails.add(detail);
            });
            outputBill.setOutputBillDetails(outputBillDetails);
            auditDeductionInventory(outputBill);
        }else {
            WarehouseOutputBill outputBill = outputBillMapper.getByBillNo(dto.getBillNo());
            outputBill.setStatus(dto.getStatus());
            outputBillMapper.updateById(outputBill);
            dto.getDetailsDtos().stream().forEach(entity->{
                WarehouseOutputBillDetail detail = detailService.getByBillMaterialCode(dto.getBillNo(),entity.getSku());
                detail.setStatus(detail.getStatus());
                detailService.updateById(detail);
            });
        }
    }


    @Override
    public void auditDeductionAtomic(WarehouseOutputBill outputBill) {
        RLock lock = RedissonUtils.getRLock(redissonClient, WarehouseConstant.UPDATE_INVENTORY_LOCK_KEY);
        boolean isLock = RedissonUtils.getRLockExpireTime(lock, 6L, 6L);
        //出库单明细出库。防止并发修改库存，加锁
        if (isLock) {
            try {
                lock.lock();
                warehouseOutputBillService.auditDeduction(outputBill);
            } finally {
                if (lock != null && lock.isLocked() && lock.isHeldByCurrentThread()) {
                    lock.unlock();
                }
            }
        } else {
            throw new BusinessException("有其他人正在操作，请稍后再试！");
        }
    }

    /**
     * 直接出库，无需审核通过，原子扣减库存
     **/
    @Override
    public void autoDeductionAtomic(WarehouseOutputBill outputBill) {
//        RLock lock = RedissonUtils.getRLock(redissonClient, WarehouseConstant.UPDATE_INVENTORY_LOCK_KEY);
//        boolean isLock = RedissonUtils.getRLockExpireTime(lock, 6L, 5L);
//        //出库单单明细出库。防止并发修改库存，加锁
//        if (isLock) {
//            try {
//                lock.lock();
//                warehouseOutputBillService.autoDeduction(outputBill);
//            } finally {
//                if (lock != null && lock.isLocked() && lock.isHeldByCurrentThread()) {
//                    lock.unlock();
//                }
//            }
//        } else {
//            throw new BusinessException("有其他人正在操作，请稍后再试！");
//        }
        warehouseOutputBillService.autoDeduction(outputBill);
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public void autoDeduction(WarehouseOutputBill outputBill) {
        /**
         * 直接出库无需审核，且非手动创建的出库单，要自动扣减库存，则没有审核过程中选择具体物料库存的步骤，也就没有唯一码、批次号、库位等信息，只有物料编码和需求量
         * 假如某个物料有多条库存数据，则自动扣减时，是不知道扣减哪个库存数据的。
         * 此时，采用随机扣减的原则
         **/
        WarehouseOutputBill byBillNo = this.getByBillNo(outputBill.getBillNo());
        if (byBillNo == null) {
            throw new BusinessException("出库单号【" + outputBill.getBillNo() + "】信息为空，自动出库失败！");
        }
        autoDeductionInventory(outputBill);
        //TODO wupeng 此处需要调用其他需要反馈的系统接口
        warehouseConsumerService.afterAutoOutput(outputBill);
    }

    /**
     * 取消出库申请
     **/
    @Override
    @Transactional(rollbackFor = Exception.class)
    public void cancelApply(String billNo) {
        if (StringUtils.isBlank(billNo)) {
            throw new BusinessException("出库单号为空！");
        }
        WarehouseOutputBill byBill = warehouseOutputBillService.getByBillNo(billNo);
        if (byBill == null) {
            throw new BusinessException("该单据不存在！");
        }
        String status = byBill.getStatus();
        if (!OutputBillStatusEnum.待出库.getValue().equals(status)) {
            throw new BusinessException("该单据状态为【" + status + "】，无法取消！");
        }
        byBill.setStatus(InputBillStatusEnum.已取消.getValue());
        byBill.setUpdateUserName(BaseContextHandler.getUsername());
        byBill.setUpdateUserId(BaseContextHandler.getUserID());
        byBill.setUpdateTime(new Date());
        //更新出库单状态
        warehouseOutputBillService.saveOrUpdate(byBill);
        //更新出库明细状态
        List<WarehouseOutputBillDetail> billDetails = detailService.getByBillNo(billNo);
        if (billDetails == null || billDetails.size() == 0) {
            throw new BusinessException("该单据明细为空，取消失败！");
        }
        billDetails.forEach(e -> {
            e.setStatus(OutputBillStatusEnum.已取消.getValue());
            e.setUpdateUserName(BaseContextHandler.getUsername());
            e.setUpdateUserId(BaseContextHandler.getUserID());
            e.setUpdateTime(byBill.getUpdateTime());
        });
        detailService.saveOrUpdateBatch(billDetails);
        //TODO wupeng 此处需要调用其他需要反馈的系统接口
        warehouseConsumerService.afterOutputCancel(byBill);
    }

    /**
     * 拒绝出库
     **/
    @Override
    public void rejectApply(WarehouseOutputBill outputBill) {
        String billNo = outputBill.getBillNo();
        if (StringUtils.isBlank(billNo)) {
            throw new BusinessException("单据号不能为空！");
        }
        WarehouseOutputBill byBillNo = this.getByBillNo(billNo);
        if (byBillNo == null) {
            throw new BusinessException("该单据不存在！");
        }
        List<WarehouseOutputBillDetail> billDetails = detailService.getByBillNo(billNo);
        if (billDetails == null || billDetails.size() == 0) {
            throw new BusinessException("该单据明细为空，拒收失败！");
        }
        String status = byBillNo.getStatus();
        if (!OutputBillStatusEnum.待出库.getValue().equals(status)) {
            throw new BusinessException("该单据已经是【" + status + "】状态，无法拒收！");
        }
        byBillNo.setStatus(InputBillStatusEnum.拒收.getValue());
        byBillNo.setUpdateUserName(BaseContextHandler.getUsername());
        byBillNo.setUpdateUserId(BaseContextHandler.getUserID());
        byBillNo.setUpdateTime(new Date());
        //更新入库单状态
        warehouseOutputBillService.saveOrUpdate(byBillNo);
        //更新入库明细状态
        billDetails.forEach(e -> {
            e.setStatus(OutputBillStatusEnum.拒绝.getValue());
            e.setUpdateUserName(BaseContextHandler.getUsername());
            e.setUpdateUserId(BaseContextHandler.getUserID());
            e.setUpdateTime(byBillNo.getUpdateTime());
        });
        detailService.saveOrUpdateBatch(billDetails);
        //TODO wupeng 此处需要调用其他需要反馈的系统接口
        warehouseConsumerService.afterOutputReject(byBillNo);
    }

    @Override
    public WarehouseOutputBill getByBillNo(String billNo) {
        return outputBillMapper.getByBillNo(billNo);
    }

    /**
     * 审核扣减库存，更新出库单、出库单明细状态
     **/
    private void auditDeductionInventory(WarehouseOutputBill outputBill) {
        //修改出库单状态为 已出库
        this.updateOutputBill(outputBill, OutputBillStatusEnum.已出库);
        //保存审核记录单
        WarehouseOutputRecordBill warehouseOutputRecordBill = this.saveRecordBill(outputBill);
        //数据库中的出库明细，即出库申请单的申请明细
        List<WarehouseOutputBillDetail> outputBillDetails = detailService.getByBillId(outputBill.getId());
        if (CollectionUtils.isEmpty(outputBillDetails)) {
            throw new BusinessException("出库明细为空，请检查数据");
        }
        //更新物料明细状态
        this.updateDetail(outputBillDetails, OutputBillStatusEnum.已出库);
        //前端传递的真正审核的明细，需要审核的单据，申请明细和审核明细可能是不一样的，一条申请单明细可能需要多个库位出库，就会产生多条审核明细
        List<WarehouseOutputBillDetail> outputBillDetailsDtos = outputBill.getOutputBillDetails();
        //审核明细

        for (WarehouseOutputBillDetail billDetail : outputBillDetailsDtos) {
            String materialModelCode = billDetail.getMaterialModelCode();
            if (StringUtils.isBlank(materialModelCode)) {
                throw new BusinessException("出库的物料编码为空，请检查数据");
            }
            Long warehouseId = billDetail.getWarehouseId();
            if (warehouseId == null || warehouseId == 0) {
                throw new BusinessException("出库明细的库位信息为空，请检查数据");
            }
            String uniqueCode = billDetail.getUniqueCode();
            String batchNo = billDetail.getBatchNo();
            BigDecimal outputCount = billDetail.getOutputCount(); //审核出库时，出库人填写的真实要出库的数量
            BigDecimal requireCount = billDetail.getRequireCount(); //需求数量，是创建出库单时，需要的数量
            if (outputCount == null || outputCount.compareTo(new BigDecimal("0")) == 0) {
                //如果出库数量为空，则默认出库数量为需求数量(正常情况下，二者是相等的)
                outputCount = requireCount;
            }
            if (requireCount == null || requireCount.compareTo(new BigDecimal("0")) == 0) {
                requireCount = outputCount;
            }
            if (StringUtils.isNotBlank(uniqueCode)) {
                WarehouseInventory byUniqueCode = inventoryService.getByUniqueCode(uniqueCode);
                if (byUniqueCode == null) {
                    throw new BusinessException("唯一码【" + uniqueCode + "】对应的物料库存不足，请重新选择！");
                }
                //唯一码物料库存只能是 1
                if (outputCount.compareTo(new BigDecimal("1")) != 0) { //出库数量是不是1
                    throw new BusinessException("唯一码【" + uniqueCode + "】物料出库数量只能为1，请检查数据！");
                }
                BigDecimal lockQuantity = byUniqueCode.getLockQuantity() == null ? new BigDecimal("0") : byUniqueCode.getLockQuantity();
                BigDecimal quantity = byUniqueCode.getQuantity();
                //剩余库存为1时可以出库，扣减库存为0，并删除该条库存数据
                BigDecimal remain = quantity.subtract(lockQuantity);
                if (remain.compareTo(new BigDecimal("1")) == 0) {
                    byUniqueCode.setQuantity(new BigDecimal("0"));
                    byUniqueCode.setDel(true);
                    inventoryService.updateInventory(byUniqueCode, 2);
                } else {
                    throw new BusinessException("唯一码【" + uniqueCode + "】对应的物料库存不足，请重新选择！");
                }
            } else {
                //根据 唯一码为空、批次号、物料、库位  查询物料库存数据
                List<WarehouseInventory> warehouseInventory = inventoryService.getByMaterialAndBatchNoAndWarehouseId(batchNo, billDetail.getMaterialModelCode(), billDetail.getWarehouseId());
                warehouseInventory = warehouseInventory.stream().filter(e -> StringUtils.isBlank(e.getUniqueCode())).collect(Collectors.toList());
                if (CollectionUtils.isEmpty(warehouseInventory)) {
                    throw new BusinessException("批次号【" + batchNo + "】，物料名称【" + billDetail.getMaterialModelName() + "】，库位名称【" + billDetail.getWarehouseId() + "】对应的库存数据不存在，请检查数据！");
                }
                WarehouseInventory inventory = warehouseInventory.get(0);
                BigDecimal lockQuantity = inventory.getLockQuantity() == null ? new BigDecimal("0") : inventory.getLockQuantity();
                BigDecimal quantity = inventory.getQuantity();
                BigDecimal remain = quantity.subtract(lockQuantity); //剩余库存
                if (outputCount.compareTo(remain) == 1) {
                    //出库数量大于剩余库存
                    throw new BusinessException("批次号【" + batchNo + "】，物料名称【" + billDetail.getMaterialModelName() + "】，库位名称【" + billDetail.getWarehouseId() + "】对应的库存不足，请检查数据！");
                }
                quantity = quantity.subtract(outputCount);
                if (quantity.compareTo(new BigDecimal("0")) == 0) {
                    inventory.setDel(true);
                }
                inventory.setQuantity(quantity);
                inventoryService.updateInventory(inventory, 2);
            }
            //新增审核记录明细
            this.saveRecordBillDetail(warehouseOutputRecordBill, billDetail, requireCount, outputCount, null, OutputBillStatusEnum.已出库);
        }
    }

    /**
     * 自动扣减库存，更新出库单、出库单明细状态
     **/
    private void autoDeductionInventory(WarehouseOutputBill outputBill) {
        //修改出库单状态为 已出库
        this.updateOutputBill(outputBill, OutputBillStatusEnum.已出库);
        //保存审核记录单
        WarehouseOutputRecordBill warehouseOutputRecordBill = this.saveRecordBill(outputBill);
        //出库申请单的申请明细。当一条申请明细需要从多个库位扣减库存时，对应多条出库审核明细
        List<WarehouseOutputBillDetail> outputBillDetails = outputBill.getOutputBillDetails();
        if (CollectionUtils.isEmpty(outputBillDetails)) {
            throw new BusinessException("出库明细为空，请检查数据");
        }
        //更新物料明细状态
        this.updateDetail(outputBillDetails, OutputBillStatusEnum.已出库);
        for (WarehouseOutputBillDetail billDetail : outputBillDetails) {
            String materialModelCode = billDetail.getMaterialModelCode();
            if (StringUtils.isBlank(materialModelCode)) {
                throw new BusinessException("出库的物料编码为空，请检查数据");
            }
            String uniqueCode = billDetail.getUniqueCode();
            String batchNo = billDetail.getBatchNo();
            BigDecimal outputCount = billDetail.getOutputCount(); //审核出库时，出库人填写的真实要出库的数量
            BigDecimal requireCount = billDetail.getRequireCount(); //需求数量，是创建出库单时，需要的数量
            if (outputCount == null || outputCount.compareTo(new BigDecimal("0")) == 0) {
                //如果出库数量为空，则默认出库数量为需求数量(正常情况下，二者是相等的)
                outputCount = requireCount;
            }
            if (requireCount == null || requireCount.compareTo(new BigDecimal("0")) == 0) {
                requireCount = outputCount;
            }
            if (StringUtils.isNotBlank(uniqueCode)) {
                WarehouseInventory byUniqueCode = inventoryService.getByUniqueCode(uniqueCode);
                if (byUniqueCode == null) {
                    throw new BusinessException("唯一码【" + uniqueCode + "】对应的物料库存不足，请重新选择！");
                }
                //唯一码物料库存只能是 1
                if (outputCount.compareTo(new BigDecimal("1")) != 0) { //出库数量是不是1
                    throw new BusinessException("唯一码【" + uniqueCode + "】物料出库数量只能为1，请检查数据！");
                }
                BigDecimal lockQuantity = byUniqueCode.getLockQuantity() == null ? new BigDecimal("0") : byUniqueCode.getLockQuantity();
                BigDecimal quantity = byUniqueCode.getQuantity();
                //剩余库存为1时可以出库，扣减库存为0，并删除该条库存数据
                BigDecimal remain = quantity.subtract(lockQuantity);
                if (remain.compareTo(new BigDecimal("1")) != 0) {
                    throw new BusinessException("唯一码【" + uniqueCode + "】对应的物料库存不足，请重新选择！");
                }
                byUniqueCode.setQuantity(new BigDecimal("0"));
                byUniqueCode.setDel(true);
                inventoryService.updateInventory(byUniqueCode, 2);
                //新增审核明细
                this.saveRecordBillDetail(warehouseOutputRecordBill, billDetail, requireCount, outputCount, byUniqueCode, OutputBillStatusEnum.已出库);
            } else {
                Long warehouseId = billDetail.getWarehouseId();
                if (warehouseId != null && warehouseId != 0) {
                    List<WarehouseInventory> warehouseInventories = inventoryService.getByMaterialAndBatchNoAndWarehouseId(batchNo, materialModelCode, warehouseId);
                    warehouseInventories = warehouseInventories.stream().filter(e -> StringUtils.isBlank(e.getUniqueCode())).collect(Collectors.toList());
                    if (CollectionUtils.isEmpty(warehouseInventories)) {
                        throw new BusinessException("批次号【" + batchNo + "】，物料名称【" + billDetail.getMaterialModelName() + "】，库位名称【" + billDetail.getWarehouseId() + "】对应的库存数据不存在，请检查数据！");
                    }
                    WarehouseInventory inventory = warehouseInventories.get(0);
                    BigDecimal quantity = inventory.getQuantity();
                    BigDecimal lockQuantity = inventory.getLockQuantity() == null ? new BigDecimal("0") : inventory.getLockQuantity();
                    BigDecimal remain = quantity.subtract(lockQuantity);
                    if (outputCount.compareTo(remain) == 1) {
                        //出库数量大于库存数量
                        throw new BusinessException("批次号【" + batchNo + "】，物料名称【" + billDetail.getMaterialModelName() + "】，库位名称【" + billDetail.getWarehouseId() + "】对应的库存不足，请检查数据！");
                    }
                    //出库数量小于库存数量，出库成功，更新库存
                    quantity = quantity.subtract(outputCount);
                    if (quantity.compareTo(new BigDecimal("0")) == 0) {
                        inventory.setDel(true);
                    }
                    inventory.setQuantity(quantity);
                    inventoryService.updateInventory(inventory, 2);
                    //新增审核明细
                    this.saveRecordBillDetail(warehouseOutputRecordBill, billDetail, requireCount, outputCount, inventory, OutputBillStatusEnum.已出库);
                } else {
                    //当出库明细中未填写库位信息，则存在多条库存数据时，随机扣减库存，此时会产生多条审核记录明细
                    List<WarehouseInventory> inventories = inventoryService.getByMaterialAndBatchNoAndWarehouseId(batchNo, materialModelCode, null);
                    inventories = inventories.stream().filter(e -> StringUtils.isBlank(e.getUniqueCode())).collect(Collectors.toList());
                    if (CollectionUtils.isEmpty(inventories)) {
                        throw new BusinessException("物料编码【" + materialModelCode + "】暂无库存记录，请检查数据！");
                    }
                    BigDecimal stillNeeded = outputCount; //还需要的出库数量
                    for (int i = 0; i < inventories.size(); i++) {
                        if (stillNeeded.compareTo(new BigDecimal("0")) == 0) {
                            break;
                        }
                        WarehouseInventory inventory = inventories.get(i);
                        BigDecimal quantity = inventory.getQuantity();
                        BigDecimal lockQuantity = inventory.getLockQuantity() == null ? new BigDecimal("0") : inventory.getLockQuantity();
                        BigDecimal remain = quantity.subtract(lockQuantity); //剩余库存
                        BigDecimal truthOutputCount = outputCount;
                        if (remain.compareTo(stillNeeded) == -1) {
                            //如果库存数量小于需求数量，删除当前扣完的库存数据，继续扣减下一个库存
                            quantity = lockQuantity;
                            stillNeeded = outputCount.subtract(remain);
                            truthOutputCount = remain;
                        } else {
                            //如果剩余库存数量大于或等于需求数量，更新库存
                            quantity = quantity.subtract(stillNeeded);
                            truthOutputCount = stillNeeded;
                            stillNeeded = new BigDecimal("0");
                        }
                        if (quantity.compareTo(new BigDecimal("0")) == 0) {
                            inventory.setDel(true);
                        }
                        inventory.setQuantity(quantity);
                        inventoryService.updateInventory(inventory, 2);
                        //随机扣减库存，如果扣减了多个库位的库存，应该保存多条审核明细
                        this.saveRecordBillDetail(warehouseOutputRecordBill, billDetail, requireCount, truthOutputCount, inventory, OutputBillStatusEnum.已出库);
                    }
                    if (stillNeeded.compareTo(new BigDecimal("0")) == 1) {
                        throw new BusinessException("物料编码【" + materialModelCode + "】库存不足，请检查数据！");
                    }
                }
            }
        }
    }

    private void updateOutputBill(WarehouseOutputBill outputBill, OutputBillStatusEnum statusEnum) {
        outputBill.setStatus(statusEnum.getValue());
        outputBill.setUpdateTime(new Date());
        outputBill.setUpdateUserId(BaseContextHandler.getUserID());
        outputBill.setUpdateUserName(BaseContextHandler.getUsername());
        warehouseOutputBillService.saveOrUpdate(outputBill);
    }

    private void updateDetail(List<WarehouseOutputBillDetail> outputBillDetails, OutputBillStatusEnum statusEnum) {
        outputBillDetails.forEach(billDetail -> {
            billDetail.setStatus(statusEnum.getValue());
            billDetail.setUpdateTime(new Date());
            billDetail.setUpdateUserId(BaseContextHandler.getUserID());
            billDetail.setUpdateUserName(BaseContextHandler.getUsername());
        });
        detailService.saveOrUpdateBatch(outputBillDetails);
    }

    private WarehouseOutputRecordBill saveRecordBill(WarehouseOutputBill outputBill) {
        WarehouseOutputRecordBill outputRecordBill = modelMapper.map(outputBill, WarehouseOutputRecordBill.class);
        outputRecordBill.setId(null);
        outputRecordBill.setRecordBillNo(CodeGenUtil.genOutputRecordBillNo());
        outputRecordBill.setOutputBillId(outputBill.getId());
        outputRecordBill.setOutputBillNo(outputBill.getBillNo());
        if (ObjectUtils.isEmpty(outputBill.getCreateTime())) {
            outputRecordBill.setCreateTime(new Date());
        }
        if (ObjectUtils.isEmpty(outputBill.getCreateUserName())) {
            outputRecordBill.setCreateUserName(BaseContextHandler.getUsername());
        }
        if (ObjectUtils.isEmpty(outputBill.getCreateUserId())) {
            outputRecordBill.setCreateUserId(BaseContextHandler.getUserID());
        }
        outputRecordBillService.saveOrUpdate(outputRecordBill);
        return outputRecordBill;
    }

    private void saveRecordBillDetail(WarehouseOutputRecordBill outputRecordBill, WarehouseOutputBillDetail outputBillDetail,
                                      BigDecimal requireCount, BigDecimal outputCount, WarehouseInventory inventory, OutputBillStatusEnum statusEnum) {
        WarehouseOutputRecordBillDetail outputRecordBillDetail = modelMapper.map(outputBillDetail, WarehouseOutputRecordBillDetail.class);
        outputRecordBillDetail.setId(null);
        outputRecordBillDetail.setRecordBillNo(outputRecordBill.getRecordBillNo());
        outputRecordBillDetail.setRecordBillId(outputRecordBill.getId());
        outputRecordBillDetail.setOutputBillId(outputBillDetail.getOutputBillId());
        outputRecordBillDetail.setOutputBillNo(outputBillDetail.getOutputBillNo());
        outputRecordBillDetail.setRequireCount(requireCount);
        outputRecordBillDetail.setOutputCount(outputCount);
        outputBillDetail.setOutputCount(outputCount);
        if (inventory != null) {
            outputRecordBillDetail.setWarehouseId(inventory.getWarehouseId());
            outputRecordBillDetail.setWarehouseName(inventory.getWarehouseName());
            outputRecordBillDetail.setWarehouseCode(inventory.getWarehouseCode());
        }
        outputRecordBillDetail.setStatus(statusEnum.getValue());
        if (ObjectUtils.isEmpty(outputRecordBill.getCreateTime())) {
            outputRecordBillDetail.setCreateTime(new Date());
        }
        if (ObjectUtils.isEmpty(outputRecordBill.getCreateUserName())) {
            outputRecordBillDetail.setCreateUserName(BaseContextHandler.getUsername());
        }
        if (ObjectUtils.isEmpty(outputRecordBill.getCreateUserId())) {
            outputRecordBillDetail.setCreateUserId(BaseContextHandler.getUserID());
        }
        outputRecordBillDetailService.saveOrUpdate(outputRecordBillDetail);
    }

}